#include <gcc-plugin.h>
#include <plugin-version.h>
#include <tree.h>
#include <basic-block.h>
#include <gimple.h>
#include <tree-pass.h>
#include <context.h>
#include <function.h>
#include <gimple-iterator.h>
#include <gimple-pretty-print.h>

/******************************/
/****   TD2 - QUESTION 7   ****/
/******************************/

/* Build a filename (as a string) based on function name */
static char *
cfgviz_generate_filename(function *fun, const char *suffix)
{
	char *target_filename;

	target_filename = (char *)xmalloc(2048 * sizeof(char));

	snprintf(target_filename, 1024, "%s_%s_%d_%s.dot",
			 current_function_name(),
			 LOCATION_FILE(fun->function_start_locus),
			 LOCATION_LINE(fun->function_start_locus),
			 suffix);

	return target_filename;
}

/* Dump the graphviz representation of function 'fun' in file 'out' */
static void
cfgviz_internal_dump(function *fun, FILE *out)
{
	// Print the header line and open the main graph
	fprintf(out, "Digraph G{\n");

	basic_block bb;
	FOR_ALL_BB_FN(bb, cfun)
	{
		fprintf(out, "N%d[label = \" Node %d:\\n", bb->index, bb->index);
		gimple_stmt_iterator gsi;
		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi))
		{
			print_gimple_stmt(out, gsi_stmt(gsi), 0, TDF_LINENO);
		}
		fprintf(out, "\" shape = box]\n");
	}

	FOR_ALL_BB_FN(bb, cfun)
	{
		edge e;
		edge_iterator ei;
		FOR_EACH_EDGE(e, ei, bb->succs)
		{
			fprintf(out, "N%d -> N%d\n", bb->index, e->dest->index);
		}
	}

	// Close the main graph
	fprintf(out, "}\n");
}

void cfgviz_dump(function *fun, const char *suffix)
{
	char *target_filename;
	FILE *out;

	target_filename = cfgviz_generate_filename(fun, suffix);

	printf("[GRAPHVIZ] Generating CFG of function %s in file <%s>\n",
		   current_function_name(), target_filename);

	out = fopen(target_filename, "w");

	cfgviz_internal_dump(fun, out);

	fclose(out);
	free(target_filename);
}

/******************************/
/**   TD2 - FIN QUESTION 7   **/
/******************************/

int plugin_is_GPL_compatible;

const pass_data my_pass_data = {
	GIMPLE_PASS,   /* type */
	"MY_PASS",	   /* name */
	OPTGROUP_NONE, /* optinfo_flags */
	TV_OPTIMIZE,   /* tv_id */
	0,			   /* properties_required */
	0,			   /* properties_provided */
	0,			   /* properties_destroyed */
	0,			   /* todo_flags_start */
	0,			   /* todo_flags_finish */
};

class my_pass : public gimple_opt_pass
{
public:
	my_pass(gcc::context *ctxt)
		: gimple_opt_pass(my_pass_data, ctxt)
	{
	}
	my_pass *clone() { return new my_pass(g); }
	bool gate(function *fun) { return true; }
	unsigned int execute(function *fun)
	{
		cfgviz_dump(fun, "graph");
		return 0;
	}
};

int plugin_init(
	struct plugin_name_args *plugin_info,
	struct plugin_gcc_version *version)
{
	struct register_pass_info pass_info;
	my_pass p(g);
	pass_info.pass = &p;
	pass_info.reference_pass_name =
		"cfg";
	pass_info.ref_pass_instance_number = 0;
	pass_info.pos_op = PASS_POS_INSERT_AFTER;
	register_callback(
		plugin_info->base_name,
		PLUGIN_PASS_MANAGER_SETUP,
		NULL,
		&pass_info);
	return 0;
}