#include <gcc-plugin.h>
#include <plugin-version.h>
#include <tree.h>
#include <basic-block.h>
#include <gimple.h>
#include <tree-pass.h>
#include <context.h>
#include <function.h>
#include <gimple-iterator.h>
#include <bitmap.h>

#include <queue>

/* Global variable required for plugin to execute */
int plugin_is_GPL_compatible;

/* Global object (const) to represent my pass */
const pass_data my_pass_data =
	{
		GIMPLE_PASS,   /* type */
		"NEW_PASS",	   /* name */
		OPTGROUP_NONE, /* optinfo_flags */
		TV_OPTIMIZE,   /* tv_id */
		0,			   /* properties_required */
		0,			   /* properties_provided */
		0,			   /* properties_destroyed */
		0,			   /* todo_flags_start */
		0,			   /* todo_flags_finish */
};

/* Enum to represent the collective operations */
enum mpi_collective_code
{
#define DEFMPICOLLECTIVES(CODE, NAME) CODE,
#include "MPI_collectives.def"
	LAST_AND_UNUSED_MPI_COLLECTIVE_CODE
#undef DEFMPICOLLECTIVES
};

/* Name of each MPI collective operations */
#define DEFMPICOLLECTIVES(CODE, NAME) NAME,
const char *const mpi_collective_name[] = {
#include "MPI_collectives.def"
};
#undef DEFMPICOLLECTIVES

void td_isol_print(int td)
{
	printf("\n\n\n");
	printf("/****************************************************************************************************/\n");
	printf("/****************************************************************************************************/\n");
	printf("/*************************                       TD%d                        *************************/\n", td);
	printf("/****************************************************************************************************/\n");
	printf("/****************************************************************************************************/\n");
	printf("\n\n\n");
}

void function_isol_print(function *fun)
{
	printf("\n\n\n");
	printf("/************************************************************************************************************************/\n");
	printf("/************************************************************************************************************************/\n");
	printf("/***********************************            %s               ***********************************/\n", function_name(fun));
	printf("/************************************************************************************************************************/\n");
	printf("/************************************************************************************************************************/\n");
	printf("\n\n\n");
}

/****************************************************************************************************/
/****************************************************************************************************/
/*************************                       TD2                        *************************/
/****************************************************************************************************/
/****************************************************************************************************/

/******************************/
/**** TD2 - QUESTION 3 & 4 ****/
/******************************/

const char *td2_q3_q4_print_func_name(function *fun)
{
	//            const char * fname = fndecl_name(fun->decl);
	const char *fname = function_name(fun);
	//            const char * fname = current_function_name();
	printf("\t ... in function %s\n", fname);

	return fname;
}

/******************************/
/** TD2 - FIN QUESTION 3 & 4 **/
/******************************/

/******************************/
/****   TD2 - QUESTION 8   ****/
/******************************/

void td2_q8_print_called_functions(basic_block bb)
{
	gimple_stmt_iterator gsi;

	/* Iterate on gimple statements in the current basic block */
	for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi))
	{
		/* Get the current statement */
		gimple *stmt = gsi_stmt(gsi);

		if (is_gimple_call(stmt))
		{
			tree t;
			const char *callee_name;

			t = gimple_call_fndecl(stmt);
			callee_name = IDENTIFIER_POINTER(DECL_NAME(t));

			printf("          |||++|||      - gimple statement is a function call: function called is \" %s \" \n", callee_name);
		}
	}
}
/******************************/
/**   TD2 - FIN QUESTION 8   **/
/******************************/

/******************************/
/**** TD2 - QUESTION 5 & 6 ****/
/******************************/

void td2_q5_q6_print_blocks(function *fun)
{
	basic_block bb;
	gimple_stmt_iterator gsi;
	gimple *stmt;

	FOR_EACH_BB_FN(bb, fun)
	{
		gsi = gsi_start_bb(bb);
		stmt = gsi_stmt(gsi);
		printf("          |||++||| BLOCK INDEX %d : LINE %d\n", bb->index, gimple_lineno(stmt));

		td2_q8_print_called_functions(bb);
	}
}

/******************************/
/** TD2 - FIN QUESTION 5 & 6 **/
/******************************/

/****************************************************************************************************/
/****************************************************************************************************/
/*************************                     FIN TD2                      *************************/
/****************************************************************************************************/
/****************************************************************************************************/

/****************************************************************************************************/
/****************************************************************************************************/
/*************************                       TD3                        *************************/
/****************************************************************************************************/
/****************************************************************************************************/

/******************************/
/**** TD3 - QUESTION 1 à 3 ****/
/******************************/

enum mpi_collective_code td3_q1_q2_q3_is_mpi_call(gimple *stmt, int bb_index)
{
	enum mpi_collective_code returned_code;

	returned_code = LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;

	if (is_gimple_call(stmt))
	{
		tree t;
		const char *callee_name;
		int i;
		bool found = false;

		t = gimple_call_fndecl(stmt);
		callee_name = IDENTIFIER_POINTER(DECL_NAME(t));

		i = 0;
		while (!found && i < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
		{
			if (strncmp(callee_name, mpi_collective_name[i], strlen(mpi_collective_name[i])) == 0)
			{
				found = true;
				returned_code = (enum mpi_collective_code)i;
			}
			i++;
		}
	}

	if (returned_code != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
	{
		printf("    -->   |||++||| BB %d [MPI CALL] Found call to %s\n", bb_index, mpi_collective_name[returned_code]);
	}

	return returned_code;
}

/******************************/
/** TD3 - FIN QUESTION 1 à 3 **/
/******************************/

/******************************/
/****   TD3 - QUESTION 5   ****/
/******************************/

void clean_aux_field(function *fun, long val)
{
	basic_block bb;

	/* Traverse all BBs to clean 'aux' field */
	FOR_ALL_BB_FN(bb, fun)
	{
		bb->aux = (void *)val;
	}
}
/******************************/
/**   TD3 - FIN QUESTION 5   **/
/******************************/

void td3_q1_q2_q3_q4_read_and_mark_mpi(function *fun)
{
	basic_block bb;
	gimple_stmt_iterator gsi;
	gimple *stmt;

	FOR_EACH_BB_FN(bb, fun)
	{

		/* Iterate on gimple statements in the current basic block */
		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi))
		{

			/******************************/
			/**   TD3 - QUESTION 1 à 4   **/
			/******************************/
			stmt = gsi_stmt(gsi);

			enum mpi_collective_code c;
			c = td3_q1_q2_q3_is_mpi_call(stmt, bb->index);
			if (bb->aux == (void *)LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
			{
				bb->aux = (void *)c;
			}

			/******************************/
			/** TD3 - FIN QUESTION 1 à 4 **/
			/******************************/
		}
	}
}

/******************************/
/****   TD3 - QUESTION 7   ****/
/******************************/

int td3_q7_get_nb_mpi_calls_in_bb(basic_block bb)
{
	gimple_stmt_iterator gsi;
	int nb_mpi_coll = 0;

	for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi))
	{
		gimple *stmt = gsi_stmt(gsi);

		enum mpi_collective_code c;
		c = td3_q1_q2_q3_is_mpi_call(stmt, bb->index);

		if (c != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
		{
			nb_mpi_coll++;
		}
	}

	printf(" == BB %d == contains %d MPI collective(s)\n",
		   bb->index, nb_mpi_coll);

	return nb_mpi_coll;
}

bool td3_q7_check_multiple_mpi_calls_per_bb(function *fun)
{
	basic_block bb;
	bool is_multiple_mpi_coll = false;
	int nb_mpi_coll_in_bb = 0;
	FOR_EACH_BB_FN(bb, fun)
	{
		nb_mpi_coll_in_bb = td3_q7_get_nb_mpi_calls_in_bb(bb);
		if (nb_mpi_coll_in_bb > 1)
		{
			is_multiple_mpi_coll = true;
		}
	}

	return is_multiple_mpi_coll;
}

/******************************/
/**   TD3 - FIN QUESTION 7   **/
/******************************/

/******************************/
/****   TD3 - QUESTION 8   ****/
/******************************/

void td3_q8_split_multiple_mpi_calls(function *fun)
{
	basic_block bb;

	FOR_EACH_BB_FN(bb, fun)
	{
		int n = td3_q7_get_nb_mpi_calls_in_bb(bb);

		if (n > 1)
		{
			gimple_stmt_iterator gsi;

			printf("[SPLIT] Need to split BB %d (%d collectives)\n",
				   bb->index, n);
			for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi))
			{
				gimple *stmt = gsi_stmt(gsi);
				enum mpi_collective_code c;

				c = td3_q1_q2_q3_is_mpi_call(stmt, bb->index);

				if (c != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
				{
					split_block(bb, stmt);
				}
			}
		}
	}
}

/******************************/
/**   TD3 - FIN QUESTION 8   **/
/******************************/

/****************************************************************************************************/
/****************************************************************************************************/
/*************************                     FIN TD3                      *************************/
/****************************************************************************************************/
/****************************************************************************************************/

/****************************************************************************************************/
/****************************************************************************************************/
/*************************                       TD4                        *************************/
/****************************************************************************************************/
/****************************************************************************************************/

/******************************/
/****   TD4 - QUESTION 5   ****/
/******************************/

void td4_q5_q6_print_dominators(function *fun)
{
	printf("DOMINATORS:\n");

	basic_block bb;
	FOR_ALL_BB_FN(bb, fun)
	{
		vec<basic_block> dominators;
		//		dominators = get_dominated_by(CDI_DOMINATORS, bb);
		dominators = get_all_dominated_blocks(CDI_DOMINATORS, bb);

		printf("Basic block BB %d dominates :", bb->index);

		int i;
		int size = dominators.length();

		for (i = 0; i < size; i++)
		{
			/******************************/
			/****   TD4 - QUESTION 6   ****/
			/******************************/
			if (dominators[i]->index != bb->index)
			{

				printf(" { BB %d }", dominators[i]->index);
			}
			/******************************/
			/**   TD4 - FIN QUESTION 6   **/
			/******************************/
		}
		printf("\n");
	}
}

/******************************/
/**   TD4 - FIN QUESTION 5   **/
/******************************/

/******************************/
/****   TD4 - QUESTION 7   ****/
/******************************/

void td4_q7_print_postdominators(function *fun)
{
	printf("POST-DOMINATORS:\n");

	basic_block bb;
	FOR_ALL_BB_FN(bb, fun)
	{
		vec<basic_block> postdoms;
		postdoms = get_all_dominated_blocks(CDI_POST_DOMINATORS, bb);

		printf("Basic block BB %d dominates :", bb->index);

		int i;
		int size = postdoms.length();

		for (i = 0; i < size; i++)
		{
			if (postdoms[i]->index != bb->index)
			{

				printf(" { BB %d }", postdoms[i]->index);
			}
		}
		printf("\n");
	}
}

/******************************/
/**   TD4 - FIN QUESTION 7   **/
/******************************/

/****************************************************************************************************/
/****************************************************************************************************/
/*************************                     FIN TD4                      *************************/
/****************************************************************************************************/
/****************************************************************************************************/

/****************************************************************************************************/
/****************************************************************************************************/
/*************************                       TD5                        *************************/
/****************************************************************************************************/
/****************************************************************************************************/

/****************************************************************************************************/
/****************************************************************************************************/
/*************************                     FIN TD5                      *************************/
/****************************************************************************************************/
/****************************************************************************************************/

/****************************************************************************************************/
/****************************************************************************************************/
/*************************                PLUGIN GRAPHVIZ                   *************************/
/****************************************************************************************************/
/****************************************************************************************************/

/* Build a filename (as a string) based on function name */
static char *cfgviz_generate_filename(function *fun, const char *suffix)
{
	char *target_filename;

	target_filename = (char *)xmalloc(1024 * sizeof(char));

	snprintf(target_filename, 1024, "%s_%s_%d_%s.dot",
			 current_function_name(),
			 LOCATION_FILE(fun->function_start_locus),
			 LOCATION_LINE(fun->function_start_locus),
			 suffix);

	return target_filename;
}

/* Dump the graphviz representation of function 'fun' in file 'out' */
static void cfgviz_internal_dump(function *fun, FILE *out, int td)
{
	basic_block bb;

	// Print the header line and open the main graph
	fprintf(out, "Digraph G{\n");

	/******************************/
	/****   TD2 - QUESTION 7   ****/
	/******************************/

	FOR_ALL_BB_FN(bb, cfun)
	{

		//
		// Print the basic block BB, with the MPI call if necessary
		//

		/******************************/
		/****   TD3 - QUESTION 6   ****/
		/******************************/

		if (td == 3 && (long)bb->aux != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
		{
			fprintf(out,
					"%d [label=\"BB %d", bb->index, bb->index);

			gimple_stmt_iterator gsi;
			gimple *stmt;
			gsi = gsi_start_bb(bb);
			stmt = gsi_stmt(gsi);

			/* Iterate on gimple statements in the current basic block */
			for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi))
			{

				stmt = gsi_stmt(gsi);

				enum mpi_collective_code returned_code;

				returned_code = LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;

				if (is_gimple_call(stmt))
				{
					tree t;
					const char *callee_name;
					int i;
					bool found = false;

					t = gimple_call_fndecl(stmt);
					callee_name = IDENTIFIER_POINTER(DECL_NAME(t));

					i = 0;
					while (!found && i < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
					{
						if (strncmp(callee_name, mpi_collective_name[i], strlen(mpi_collective_name[i])) == 0)
						{
							found = true;
							returned_code = (enum mpi_collective_code)i;
						}
						i++;
					}
				}

				if (returned_code != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
				{
					fprintf(out, " \\n %s", mpi_collective_name[returned_code]);
				}
			}

			fprintf(out, "\" shape=ellipse]\n");
		}
		/******************************/
		/**** TD3 - FIN QUESTION 6 ****/
		/******************************/
		else
		{
			fprintf(out,
					"%d [label=\"BB %d\" shape=ellipse]\n",
					bb->index,
					bb->index);
		}

		//
		// Process output edges
		//
		edge_iterator eit;
		edge e;

		FOR_EACH_EDGE(e, eit, bb->succs)
		{
			const char *label = "";
			if (e->flags == EDGE_TRUE_VALUE)
				label = "true";
			else if (e->flags == EDGE_FALSE_VALUE)
				label = "false";

			fprintf(out, "%d -> %d [color=red label=\"%s\"]\n",
					bb->index, e->dest->index, label);
		}
	}
	/******************************/
	/**   TD2 - FIN QUESTION 7   **/
	/******************************/

	// Close the main graph
	fprintf(out, "}\n");
}

void cfgviz_dump(function *fun, const char *suffix, int td)
{
	char *target_filename;
	FILE *out;

	target_filename = cfgviz_generate_filename(fun, suffix);

	printf("[GRAPHVIZ] Generating CFG of function %s in file <%s>\n",
		   current_function_name(), target_filename);

	out = fopen(target_filename, "w");

	cfgviz_internal_dump(fun, out, td);

	fclose(out);
	free(target_filename);
}

/****************************************************************************************************/
/****************************************************************************************************/
/*************************            FIN  PLUGIN GRAPHVIZ                 *************************/
/****************************************************************************************************/
/****************************************************************************************************/

void td2_through_the_cfg(function *fun)
{
	td2_q3_q4_print_func_name(fun);
	td2_q5_q6_print_blocks(fun);
}

void td3_mpi_in_blocks(function *fun)
{
	bool do_splitting = false;

	td3_q1_q2_q3_q4_read_and_mark_mpi(fun);
	cfgviz_dump(fun, "1_mpi", /*TD*/ 3);
	do_splitting = td3_q7_check_multiple_mpi_calls_per_bb(fun);

	/******************************/
	/****   TD3 - QUESTION 8   ****/
	/******************************/

	if (do_splitting)
	{
		td3_q8_split_multiple_mpi_calls(fun);
	}

	clean_aux_field(fun, LAST_AND_UNUSED_MPI_COLLECTIVE_CODE);
	td3_q1_q2_q3_q4_read_and_mark_mpi(fun);
	cfgviz_dump(fun, "2_split", /*TD*/ 3);

	/******************************/
	/**   TD3 - FIN QUESTION 8   **/
	/******************************/
}

void compute_postdominance_frontiers(bitmap_head *frontiers)
{
	edge p;
	edge_iterator ei;
	basic_block b;
	FOR_EACH_BB_FN(b, cfun)
	{
		if (EDGE_COUNT(b->succs) >= 2)
		{
			basic_block domsb = get_immediate_dominator(CDI_POST_DOMINATORS, b);
			FOR_EACH_EDGE(p, ei, b->succs)
			{
				basic_block runner = p->dest;
				if (runner == EXIT_BLOCK_PTR_FOR_FN(cfun))
					continue;

				while (runner != domsb)
				{
					if (!bitmap_set_bit(&frontiers[runner->index], b->index))
						break;
					runner = get_immediate_dominator(CDI_POST_DOMINATORS, runner);
				}
			}
		}
	}
}

void td4_q8_compute_post_dominance_frontiers(function *fun)
{
	basic_block bb;
	bitmap_head *frontiers;

	frontiers = XNEWVEC(bitmap_head, last_basic_block_for_fn(fun));
	FOR_ALL_BB_FN(bb, fun)
	{
		bitmap_initialize(&frontiers[bb->index], &bitmap_default_obstack);
	}

	compute_postdominance_frontiers(frontiers);

	FOR_ALL_BB_FN(bb, fun)
	{
		printf("BB %d", bb->index);
		bitmap_print(stdout, &frontiers[bb->index], " post-dominance frontier : ", "\n");
		bitmap_clear(&frontiers[bb->index]);
	}
	free(frontiers);
}

void td4_dom_and_postdom(function *fun)
{
	/******************************/
	/**** TD4 - QUESTION 5 & 6 ****/
	/******************************/
	calculate_dominance_info(CDI_DOMINATORS);
	td4_q5_q6_print_dominators(fun);
	/******************************/
	/** TD4 - FIN QUESTION 5 & 6 **/
	/******************************/

	/******************************/
	/****   TD4 - QUESTION 7   ****/
	/******************************/
	calculate_dominance_info(CDI_POST_DOMINATORS);
	td4_q7_print_postdominators(fun);
	/******************************/
	/**   TD4 - FIN QUESTION 7   **/
	/******************************/

	//	td4_q8_print_dfs(fun);
	td4_q8_compute_post_dominance_frontiers(fun);
}

bitmap get_set_post_dominance(bitmap_head *set, function *fun, bitmap_head *pd)
{
	bitmap_initialize(pd, &bitmap_default_obstack);

	basic_block bb;
	FOR_ALL_BB_FN(bb, fun)
	{
		if (bitmap_bit_p(set, bb->index))
		{
			auto_vec<basic_block> pd_blocks = get_all_dominated_blocks(CDI_POST_DOMINATORS, bb);
			for (unsigned int i = 0; i < pd_blocks.length(); i++)
			{
				bitmap_set_bit(pd, pd_blocks[i]->index);
			}
		}
	}

	FOR_ALL_BB_FN(bb, fun)
	{
		edge e;
		edge_iterator ei;
		int total = 0;
		int succ_n = 0;
		FOR_EACH_EDGE(e, ei, bb->succs)
		{
			basic_block succ = e->dest;
			if (bitmap_bit_p(pd, succ->index))
				succ_n++;

			total++;
		}
		if (total > 0 && succ_n == total)
			bitmap_set_bit(pd, bb->index);
	}

	return pd;
}

bitmap get_set_post_dominance_frontier(bitmap_head *set, function *fun, bitmap_head *pd)
{
	bitmap_head pd2;
	bitmap_initialize(pd, &bitmap_default_obstack);
	get_set_post_dominance(set, fun, &pd2);

	basic_block bb;
	FOR_ALL_BB_FN(bb, fun)
	{
		edge e;
		edge_iterator ei;
		FOR_EACH_EDGE(e, ei, bb->succs)
		{
			basic_block succ = e->dest;
			if (bitmap_bit_p(&pd2, succ->index) && !bitmap_bit_p(&pd2, bb->index))
			{
				bitmap_set_bit(pd, bb->index);
			}
		}
	}

	return pd;
}

void td5_bitmap_and_pdf_it(function *fun)
{
	bitmap_head *new_cfg_edges = XNEWVEC(bitmap_head, n_basic_blocks_for_fn(fun));
	basic_block src = ENTRY_BLOCK_PTR_FOR_FN(fun);
	bitmap *visited_blocks;
	visited_blocks = XNEWVEC(bitmap, n_basic_blocks_for_fn(fun));
	edge e;
	basic_block bb;

	/* Initialize the new CFG edges bitmap */
	FOR_EACH_BB_FN(bb, fun)
	{
		bitmap_initialize(&new_cfg_edges[bb->index], &bitmap_default_obstack);
	}

	/* Initialize the visited blocks bitmap */
	bitmap_initialize(*visited_blocks, &bitmap_default_obstack);
	bitmap_set_bit(*visited_blocks, src->index);

	/* Perform a breadth-first search starting from the source block */
	std::queue<basic_block> q;
	q.push(src);
	edge_iterator ei;
	printf("efzefeff");
	while (!q.empty())
	{
		basic_block bb = q.front();
		q.pop();

		/* Add outgoing edges whose destination block has not been visited yet */
		FOR_EACH_EDGE(e, ei, bb->succs)
		{
			basic_block succ = e->dest;

			if (!bitmap_bit_p(*visited_blocks, succ->index))
			{
				printf("%d -> %d\n", bb->index, succ->index);
				bitmap_set_bit(&new_cfg_edges[bb->index], succ->index);
				bitmap_set_bit(*visited_blocks, succ->index);
				q.push(succ);
			}
		}
	}

	/* Clear the visited blocks bitmap */
	bitmap_clear(*visited_blocks);

	FOR_ALL_BB_FN(bb, fun)
	{
		printf("BB %d", bb->index);
		bitmap_print(stdout, &new_cfg_edges[bb->index], " -> arcs sortants : ", "\n");
	}

	struct super_bb
	{
		basic_block bb;
		int rank;
	};

	std::queue<super_bb> qsbb;
	super_bb sbb;
	sbb.bb = src;
	sbb.rank = 0;
	qsbb.push(sbb);

	int calls[n_basic_blocks_for_fn(fun)];
	int ranks[n_basic_blocks_for_fn(fun)];

	while (!qsbb.empty())
	{
		super_bb sbb = qsbb.front();
		qsbb.pop();

		gimple_stmt_iterator gsi;
		enum mpi_collective_code c = LAST_AND_UNUSED_MPI_COLLECTIVE_CODE;

		for (gsi = gsi_start_bb(sbb.bb); !gsi_end_p(gsi); gsi_next(&gsi))
		{
			gimple *stmt = gsi_stmt(gsi);
			c = td3_q1_q2_q3_is_mpi_call(stmt, sbb.bb->index);
			if (c != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
				break;
		}

		calls[sbb.bb->index] = c;
		if (c != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
		{
			ranks[sbb.bb->index] = sbb.rank + 1;
		}
		else
		{
			ranks[sbb.bb->index] = sbb.rank;
		}

		FOR_ALL_BB_FN(bb, fun)
		{
			if (bitmap_bit_p(&new_cfg_edges[sbb.bb->index], bb->index))
			{
				super_bb sbb2;
				sbb2.bb = bb;
				sbb2.rank = ranks[sbb.bb->index];
				qsbb.push(sbb2);
			}
		}
	}

	FOR_ALL_BB_FN(bb, fun)
	{
		if (calls[bb->index] != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
		{
			printf("BB %d", bb->index);
			printf(" -> appel %s", mpi_collective_name[calls[bb->index]]);
			printf(" -> rang : %d \n", ranks[bb->index]);
		}
	}

	// suite

	int max_rank = 0;
	for (int i = 0; i < n_basic_blocks_for_fn(fun); i++)
	{
		if (ranks[i] > max_rank)
		{
			max_rank = ranks[i];
		}
	}

	bitmap_head **sets = XNEWVEC(bitmap_head *, max_rank);

	for (int i = 0; i < max_rank; i++)
	{
		sets[i] = XNEWVEC(bitmap_head, LAST_AND_UNUSED_MPI_COLLECTIVE_CODE);
		for (unsigned int j = 0; j < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE; j++)
		{
			bitmap_initialize(&sets[i][j], &bitmap_default_obstack);
		}
	}

	FOR_ALL_BB_FN(bb, fun)
	{
		if (calls[bb->index] != LAST_AND_UNUSED_MPI_COLLECTIVE_CODE)
		{
			bitmap_set_bit(&sets[ranks[bb->index] - 1][calls[bb->index]], bb->index);
		}
	}

	printf("Ensembles :\n");
	for (int i = 0; i < max_rank; i++)
	{
		for (unsigned int j = 0; j < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE; j++)
		{
			printf("Rang %d - Appel %s : ", i + 1, mpi_collective_name[j]);
			bitmap_print(stdout, &sets[i][j], "", "\n");
		}
	}

	// Suite
	printf("---\n");

	for (int i = 0; i < max_rank; i++)
	{
		for (unsigned int j = 0; j < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE; j++)
		{
			printf("Rang %d - Appel %s -", i + 1, mpi_collective_name[j]);
			bitmap_head pd;
			get_set_post_dominance(&sets[i][j], fun, &pd);
			bitmap_print(stdout, &pd, " Post-dominance : ", "\n");
		}
	}

	printf("---\n");

	for (int i = 0; i < max_rank; i++)
	{
		for (unsigned int j = 0; j < LAST_AND_UNUSED_MPI_COLLECTIVE_CODE; j++)
		{
			printf("Rang %d - Appel %s -", i + 1, mpi_collective_name[j]);
			bitmap_head pd;
			get_set_post_dominance_frontier(&sets[i][j], fun, &pd);
			bitmap_print(stdout, &pd, " Post-dominance frontier : ", "\n");
		}
	}
}

/* My new pass inheriting from regular gimple pass */
class my_pass : public gimple_opt_pass
{
public:
	my_pass(gcc::context *ctxt)
		: gimple_opt_pass(my_pass_data, ctxt)
	{
	}

	/* opt_pass methods: */

	my_pass *clone()
	{
		return new my_pass(g);
	}

	/* Gate function (shall we apply this pass?) */
	bool gate(function *fun)
	{
		function_isol_print(fun);
		printf("plugin: gate... \n");
		td2_q3_q4_print_func_name(fun);
		return true;
	}

	/* Execute function */
	unsigned int execute(function *fun)
	{
		printf("plugin: execute...\n");

		td_isol_print(/*TD*/ 2);

		/******************************/
		/**** TD2 - QUESTION 3 à 8 ****/
		/******************************/

		td2_through_the_cfg(fun);

		/******************************/
		/** TD2 - FIN QUESTION 3 à 8 **/
		/******************************/

		/******************************/
		/****   TD2 - QUESTION 7   ****/
		/******************************/

		cfgviz_dump(fun, "0_ini", /*TD*/ 2);

		/******************************/
		/**   TD2 - FIN QUESTION 7   **/
		/******************************/

		/******************************/
		/**********   TD3   ***********/
		/******************************/

		td_isol_print(/*TD*/ 3);

		/******************************/
		/****   TD3 - QUESTION 5   ****/
		/******************************/
		clean_aux_field(fun, LAST_AND_UNUSED_MPI_COLLECTIVE_CODE);
		/******************************/
		/**   TD3 - FIN QUESTION 5   **/
		/******************************/

		td3_mpi_in_blocks(fun);

		/******************************/
		/********** FIN TD3 ***********/
		/******************************/

		/******************************/
		/**********   TD4   ***********/
		/******************************/

		td_isol_print(/*TD*/ 4);
		td4_dom_and_postdom(fun);

		/******************************/
		/********** FIN TD4 ***********/
		/******************************/

		/******************************/
		/**********   TD5   ***********/
		/******************************/

		td_isol_print(/*TD*/ 5);
		td5_bitmap_and_pdf_it(fun);

		/******************************/
		/********** FIN TD5 ***********/
		/******************************/

		/******************************/
		/****   TD3 - QUESTION 5   ****/
		/******************************/
		clean_aux_field(fun, 0);
		/******************************/
		/**   TD3 - FIN QUESTION 5   **/
		/******************************/

		free_dominance_info(CDI_POST_DOMINATORS);
		free_dominance_info(CDI_DOMINATORS);

		return 0;
	}
};

/* Main entry point for plugin */
int plugin_init(struct plugin_name_args *plugin_info,
				struct plugin_gcc_version *version)
{
	struct register_pass_info my_pass_info;

	printf("plugin_init: Entering...\n");

	/* First check that the current version of GCC is the right one */

	if (!plugin_default_version_check(version, &gcc_version))
		return 1;

	printf("plugin_init: Check ok...\n");

	/* Declare and build my new pass */
	my_pass p(g);

	/* Fill info on my pass
	 (insertion after the pass building the CFG) */
	my_pass_info.pass = &p;
	my_pass_info.reference_pass_name = "cfg";
	my_pass_info.ref_pass_instance_number = 0;
	my_pass_info.pos_op = PASS_POS_INSERT_AFTER;

	/* Add my pass to the pass manager */
	register_callback(plugin_info->base_name,
					  PLUGIN_PASS_MANAGER_SETUP,
					  NULL,
					  &my_pass_info);

	printf("plugin_init: Pass added...\n");

	return 0;
}
